/********************************************************************\

  Name:         util.c
  Created by:   Stefan Ritt

  Contents:     General purpose utility functions for use in the
                PiBeta analyzer


  $Log: util.c,v $
  Revision 1.12  2000/09/15 09:34:45  midas
  Initial revision

\********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "util.h"

/*----------------------------------------------------------------*/

/*******************************************************************
  
    Marquardt fit. N-dimensional fit to a non-linear function.

*******************************************************************/

/* maximal number of parameters */

#define MAX_PAR 20

/*----------------------------------------------------------------*/

void fderiv(double x, double param[], 
            double (*func)(double x, double param[]), 
            int n_terms, double deriv[])
/******************************************************************\

  calculation of derivations of a function "func" at point x
  in respect to its parameters param[]

\******************************************************************/
{ 
int    i;
double delta, yf;

  yf = func(x, param);

  for (i=0 ; i<n_terms ; i++)
    {
    delta = fabs(param[i])/1000;
    if (delta < 1E-12) 
      delta=1E-12;

	  param[i] += delta;
	  deriv[i] = (func(x, param) - yf) / delta;
	  param[i] -=delta;
    }
}

/*----------------------------------------------------------------------*/

double calc_chisqr(double y[], double yfit[], double weight[], 
                   int n_points, int n_free)
/******************************************************************\
  calculates chisquare for a fit yfit[] in respect to a sample 
  distribution y[]
  
  Input parameters: y[]        sample
                    yfit[]     fit
                    weight[]   statistical weights 
                    n_points   dimension of y,yfit
                    n_free     degree of freedom

  Return value:     chisquare 
\******************************************************************/
{ 
double chisq;
int    i;

  for (i=0,chisq=0 ; i<n_points ; i++)
    chisq += weight[i]*(y[i]-yfit[i])*(y[i]-yfit[i]);

  return(chisq/(double)n_free);
}

/*--------------------------------------------------------------------*/

int matinv(double array[MAX_PAR][MAX_PAR], int n_order)
/******************************************************************\

  inverts matrix array of order n_order
  returns 1 if successful, -1 if matrix is singular

\******************************************************************/
{ 
double amax,save;
int    ik[MAX_PAR],jk[MAX_PAR],k,i,j,l;

  for (k=0;k<n_order;k++)
    {
    amax=0;
label:
    for (i=k;i<n_order;i++) 
      {
      for (j=k;j<n_order;j++)
        {
	      if (fabs(amax) <= fabs(array[i][j]))
          {
	        amax=array[i][j];
	        ik[k]=i;
	        jk[k]=j;
          }
        }
      }

    if (amax==0)
      return -1;

    i=ik[k];
    if ((i-k)<0) 
      goto label;
    else if((i-k)>0)
      {
      for(j=0;j<n_order;j++)
        {
	      save=array[k][j];
	      array[k][j]=array[i][j];
	      array[i][j]=-save;
        }
      }
    j=jk[k];
    if ((j-k)<0) goto label;
    else if ((j-k)>0)
      {
      for (i=0;i<n_order;i++)
        {
	      save=array[i][k];
	      array[i][k]=array[i][j];
	      array[i][j]=-save;
        }
      }

    for (i=0 ; i<n_order ; i++)
      if (i!=k) 
        array[i][k] /= -amax;
    
    for (i=0 ; i<n_order ; i++)
      for (j=0 ; j<n_order ; j++)
	      if (i!=k && j!=k)
	        array[i][j] += array[i][k]*array[k][j];

    for(j=0;j<n_order;j++)
      if(j!=k) array[k][j]/=amax;
    array[k][k]=1/amax;
    }
  
  for (l=0 ; l<n_order ; l++)
    {
    k=n_order-l-1;
    j=ik[k];
    if (j>k)
      for (i=0;i<n_order;i++)
        {
	      save=array[i][k];
	      array[i][k]=-array[i][j];
	      array[i][j]=save;
        }

    i=jk[k];
    if (i>k)
      for (j=0;j<n_order;j++)
        {
	      save=array[k][j];
	      array[k][j]=-array[i][j];
	      array[i][j]=save;
        }
    }

  return(1);
}

/*----------------------------------------------------------------*/

int marquardt_fit(double x[],  double y[], double weight[], 
                  double (*func)(double x, double param[]), 
                  int n_points, int n_terms,
                  double a[],  double sigmaa[],
                  double *flamda,
		              double yfit[], double *chisqr)
/******************************************************************\
  
    calculates a new fit contained in yfit[] to a sample
    distribution y[] by variation of the parameters a[]
    of the fit function; it also calculates the new
    parameters a[] and  the error of the parameters sigmaa[]

    Input  :  x[] array containing x-values of sample
              y[] array containing sample
              weight[] array containing weights
              n_points: number of sample points to fit
              n_terms: number of free parameters
              a[]: array containing parameter values
              flamda: marquart faktor                    

    Output:   a[] new parameters                        
              sigmaa[] errors in parameters             
              yfit[] new fit                            
              chisqr new chisquare

    Returns: 1 on success 
             -1 if variation of parameters doesn't change chi^2

\******************************************************************/
{ 
double alpha[MAX_PAR][MAX_PAR], beta[MAX_PAR], deriv[MAX_PAR];
double array[MAX_PAR][MAX_PAR], b[MAX_PAR], chisq1, buf;
int    i,j,k;

  if (n_terms > MAX_PAR)
    printf("marquard_fit: Maximum number of parameters exceeded.\n");

  /* evaluate alpha and beta matrices */
  for (i=0 ; i<n_terms ; i++)
    {
    beta[i]=0;
    for (j=0;j<=i;j++)
      alpha[i][j]=0;
    }

  for (i=0 ; i<n_points ; i++)
    {
    fderiv(x[i], a, func, n_terms, deriv);

    for (j=0 ; j<n_terms; j++)
      {
      beta[j] += weight[i]*(y[i]-yfit[i])*deriv[j];
      for (k=0 ; k<=j ; k++)
	      alpha[j][k] += weight[i]*deriv[j]*deriv[k];
      }
    }

  for (i=0 ; i<n_terms ; i++)
    for (j=0 ; j<i ; j++)
      alpha[j][i]=alpha[i][j];

  chisq1 = *chisqr;

  /* invert modified curvature matrix to find new parameters */
  do
    {
    for (j=0 ; j<n_terms ; j++)
      {
      for (k=0 ; k<n_terms ; k++)
        {
	      buf = sqrt(alpha[j][j] * alpha[k][k]);
	      if (buf == 0)
	        return -1;

	      array[j][k] = alpha[j][k]/buf;
        }
      array[j][j] = 1 + *flamda;
      }

    if (matinv(array, n_terms) == -1)
      return -1;
    
    for (j=0 ; j<n_terms ; j++)
      {
      b[j]=a[j];
      for (k=0;k<n_terms;k++)
	      b[j] += beta[k]*array[j][k] / sqrt(alpha[j][j]*alpha[k][k]);
      }

    for(i=0 ; i<n_points ; i++)
      yfit[i] = func(x[i], b);
    
    *chisqr = calc_chisqr(y, yfit, weight, n_points, n_points-n_terms);

    if (*chisqr > chisq1) 
      *flamda *= 10;

  } while (*chisqr > chisq1 && *flamda < 1E20);

  /* evaluate parameters and uncertainties */
  for (j=0;j<n_terms;j++)
    {
    a[j] = b[j];
    sigmaa[j] = sqrt(array[j][j]/alpha[j][j]);
    }

  *flamda /= 10;

  return(0);
}

/*----------------------------------------------------------------------*/

double  fit(double x[], double y[], double weight[],
            double (*func)(double x, double param[]),
            int n_points, int n_params,
            double param[], double sigmaparam[],
            double yfit[])
/******************************************************************\
  
  fit a function "func(x, param[])" to a given data set x[]/y[]
  with poisson weighted data points.

  Input:    x[]: array containing x-values of data points
            y[]: array containing value of data points
            weight[]: array with data point weights
            func(): user-supplied fitting function
            n_points: number of data points to fit
            n_params: number of parameters
            param[]: array containing initial parameter values

  Output:   param[]: array containing fitted parameter values
            sigmaparam[]: errors in parameters             
            yfit[]: new fit                            
            chisqr: new chisquare

  Returns:  chisqr: new chisquare
            -1: if fit doesn't converge. A reason for this could
            be that the variation of a parameter doesn't affect
            the chi^2 value. This happens if the starting 
            parameters are wrong or the fit function is independent
            of a given parameter.


\******************************************************************/
{
double chisqr, chiold, flamda;
int    i, status;
  /* lamda parameter for marquardt_fit */  
  flamda = 0.001;

  /* calculate initial fit with starting parameters */
  for (i=0 ; i<n_points ; i++)
    yfit[i] = func(x[i], param);

  /* calculate starting chi sqare */
  chisqr = calc_chisqr(y, yfit, weight, n_points, n_points-n_params);

  if (chisqr == 0)
    return -1;

  /* main iteration loop */
  do
    {                                 
    chiold = chisqr;
    status = marquardt_fit(x, y, weight, func, n_points, n_params,
 	                         param, sigmaparam, &flamda, 
                           yfit, &chisqr);
    if (status == -1)
      return -1;

    if (chisqr == 0)
      return -1;

    } while(fabs(chiold-chisqr)/chisqr > 0.001);

  return chisqr;
}

/*----------------------------------------------------------------*/

double gauss_fit(double *x, double *y, int n_points, 
                 double param[3], double error[3], int guess_params)
/******************************************************************\
  
  fit a gauss function to a given data set x[]/y[]
  with poisson weighted data points.

  Input:    x[]: array containing x-values of data points
            y[]: array containing value of data points
            n_points: number of data points to fit
            param[]: array containing initial parameter values
            guess_params: guess initial parameters from data

  Output:   praram[3]: amplitude, center, fwhm of fitted gaussian.

  Returns:  chisqr: new chisquare
            -1: if fit doesn't converge. A reason for this could
            be that the variation of a parameter doesn't affect
            the chi^2 value. This happens if the starting 
            parameters are wrong or the fit function is independent
            of a given parameter.


\******************************************************************/
{
double xmax, ymax, *weight, *yfit, chisqr, sigmaparam[3];
int    i, imax;

  memset(error, 0, sizeof(error));
  weight = malloc(sizeof(double)*n_points);
  yfit   = malloc(sizeof(double)*n_points);

  /* find peak */
  ymax = 0;
  for (i=0 ; i<n_points ; i++)
    {
    if (y[i] == 0)
      weight[i] = 1;
    else
      weight[i] = 1/y[i];
    if (y[i] > ymax)
      {
      imax = i;
      ymax = y[i];
      xmax = x[i];
      }
    }

  if (ymax < 5)
    goto error;

  if (guess_params)
    {
    param[0] = ymax;
    param[1] = xmax;

    /* estimate fwhm */
    for (i=imax ; i<n_points ; )
      {
      if (y[i] < ymax/2)
        break;

      i += imax > n_points/2 ? -1 : 1;
      }

    param[2] = fabs(2*(xmax-x[i]));
    }

  chisqr = fit(x, y, weight, f_gaussian, n_points, 3, param, sigmaparam, yfit);

  /* make fwhm positive */
  param[2] = fabs(param[2]);

  /* copy errors */
  memcpy(error, sigmaparam, sizeof(double)*3);

error:

  free(weight);
  free(yfit);

  return chisqr;
}

/*----------------------------------------------------------------*/

/*******************************************************************
  
    Generic fit functions

*******************************************************************/

#define FWHM_RATIO 2.35482

double f_gaussian(double x, double param[])
/* 0: peak height, 1: center, 2: FWHM */
{
double x1;
double sigma;

  /* supress division by zero */
  if (param[2] == 0) 
    return(0);       

  sigma = param[2]/FWHM_RATIO;

  x1 = (x - param[1]) / sigma;

  /* supress overflow */
  if ((x1>4) || (x1<-4)) 
    return(0);           

  /* sqare argument */
  x1 *= x1;  

  return param[0] * exp( -0.5*x1 );
}

/*----------------------------------------------------------------*/

double f_2gaussian(double x, double param[])
/* 0: peak height, 1: center, 2: FWHM of 1st gaussian
   3: peak height, 4: center, 5: FWHM of 2nd gaussian */
{
  return f_gaussian(x, param) + f_gaussian(x, param+3);
}

/*----------------------------------------------------------------*/

double f_gaussexp(double x, double param[])
/* 0: peak height, 1: center, 2: FWHM, 3: split point */
{
double x1;
double sigma;

  /* supress division by zero */
  if (param[2] == 0) 
    return(0);       

  sigma = param[2]/FWHM_RATIO;

  if ( x <= param[1]+param[3] && param[3] > 0 ||
       x >= param[1]+param[3] && param[3] <= 0)
    {
    x1 = (x - param[1]) / sigma;

    /* supress overflow */
    if ((x1>4) || (x1<-4)) 
      return(0);           
    
    x1 *= x1;
  
    return param[0] * exp( -0.5*x1 );
    }
  else
    {
    x1 = param[3]/sigma/sigma*(0.5*param[3] - (x-param[1]));
    
    /* supress underflow of exp function */
    if (x1<-20) 
      return(0); 
    if (x1>20) 
      x1 = 20;     /* overflow */
    x1 = exp(x1);

    return param[0] * x1;
    }
}

/*----------------------------------------------------------------*/

double f_trapezoid(double x, double param[])
/* 0: gauss height, 1: gauss center, 2: gauss FWHM, 
   3: top left edge, 4: top right edge, 5: trap height,
   6: slope left, 7: slope right */
{ 
double x1;

  /* supress division by zero */
  if (param[2] == 0) 
    return 0;       

  /* (x-center)/sigma */
  x1 = (x - param[1]) / (param[2]/FWHM_RATIO);    

  /* supress overflow */
  if ((x1>4) || (x1<-4)) 
    return 0;           

  /* x^2 */
  x1 *= x1;  
  

  if (x >= param[3] && x <= param[4])
    return param[5] + param[0]*exp(-0.5*x1);

  if (x >= param[4] && x <= param[4] - param[5]/param[7])
    return param[7]*x + param[5] - param[7]*param[4] + param[0]*exp(-0.5*x1);

  if (x <= param[3] && x >= (param[6]*param[3]-param[5])/param[6])
    return param[6]*x + param[5] - param[6]*param[3] + param[0] * exp(-0.5*x1);

  if (x <= (param[6]*param[3]-param[5])/param[6])
    return param[0]*exp(-0.5*x1);
  
  if (x >= param[4] - param[5]/param[7])
    return param[0]*exp(-0.5*x1);

  return 0;
}

/*******************************************************************
  
    Gain match two histos

*******************************************************************/

/*--------- Chi^2 calculation for gain_match ----------------------*/

float calc_chisq(float *adc_histo, float *ref_histo, float norm_ref, 
                  float g, int n, float x_low, float x_high, 
                  int i_left, int i_right)
{
int    i, j;
float chisq, frac, adc, ref, delta, x;

  /* calculate chi^2 for gain g */
  chisq = 0.f;

  for (i=i_left ; i<=i_right ; i++)
    {
    /* scale adc histo with g, interpolate between bins */
    x = (float) i/n*(x_high-x_low) + x_low + 0.5f;
    x /= g;

    x = (x-x_low)/(x_high-x_low)*n;
    j = (int) x;
    frac = x - j;

    if (j>=n-1 || j<0)
      adc = 0.f;
    else
      adc = adc_histo[j] + frac * ((float) adc_histo[j+1] - adc_histo[j]);

    /* normalize events */
    adc /= g;

    ref = ref_histo[i]*norm_ref;
    delta = adc - ref;

    if (ref == 0)
      chisq += delta*delta;
    else
      chisq += delta*delta/(float)fabs(ref);
    }

  chisq /= (i_right-i_left);

  return chisq;
}

/*--------- Histogram gain evaluation -----------------------------*/

void gain_match(float *ref_histo, float *adc_histo, int n,
                float x_low, float x_high,
                float roi_low, float roi_high,
                float range_low, float range_high,
                float *gain, float *error)
/******************************************************************\
  
  Evaluate gain by matching a histo in a given region to a reference
  histo.
  
  Input:    ref_histo[]    Reference histo
            adc_histo[]    Histo which is gain matched
            x_low,x_high   Physical histo limits
            roi_low,roi_high Region where to match histos
            range_low/high Gain range to check (like 0.8..1.2)

  Output:   gain           Gain
            error          Error of gain


\******************************************************************/
{
int   i, i_low, i_high;
float sum_ref, sum_adc, norm_ref, g, g1;
float chisq, min_chisq, old_chisq;

  i_low  = (int) ((roi_low-x_low)/(x_high-x_low)*n);
  i_high = (int) ((roi_high-x_low)/(x_high-x_low)*n);

  /* normalize reference histogram */
  for (i=i_low,sum_ref=0.f,sum_adc=0.f ; i<i_high && i<n ; i++)
    {
    sum_ref += ref_histo[i];
    sum_adc += adc_histo[i];
    }

  /* don't continue for empty histos */
  if (sum_ref <= 0 || sum_adc <= 0 || n<2)
    {
    *gain = 1.f;
    *error = 1.f;
    return;
    }

  /* do range check of limits */
  if (roi_low<x_low || roi_low>=roi_high ||
      roi_high>x_high)
    {
    *gain = 1.f;
    *error = 1.f;
    return;
    }

  norm_ref = sum_adc/sum_ref;

  /* vary gain inside range in steps of 1/n */
  for (g=range_low ; g<range_high ; g+=(float) 1/n)
    {
    /* calculate chi^2 for gain g */
    chisq = calc_chisq(adc_histo, ref_histo, norm_ref, g, n, 
                       x_low, x_high, i_low, i_high);

    /* check for minimal chisq */
    if (g==range_low || chisq < min_chisq)
      {
      *gain = g;
      min_chisq = chisq;
      }
    }

  /* error estimation */
  old_chisq = min_chisq;
  for (g= *gain ; g<range_high ; g+=(float) 1/n)
    {
    /* calculate chi^2 for gain g */
    chisq = calc_chisq(adc_histo, ref_histo, norm_ref, g, n, 
                       x_low, x_high, i_low, i_high);

    /* check for chisq increase of one */
    if (chisq >= min_chisq+1)
      {
      g1 = g - (chisq - (min_chisq+1))/(chisq-old_chisq)/n;
      *error = g1 - *gain;
      break;
      }

    old_chisq = chisq;
    }
}

/*--------- raytrace for tomography like ray tracing --------------*/

void raytrace(int nx, double size_x, int ny, double size_y, int nz, double size_z,
              double x1, double y1, double z1, double x2, double y2, double z2,
              int *n_hit, int *array_i, int *array_j, int *array_k,
              double *array_pathlength)
/********************************************************************\

  Routine: raytrace

  Purpose: Traces a ray through a box with small cubic cells.

           The box consists of nx*ny*nz cells each size_x*size_y*
           size_z large.

           The ray begins at x1/y1/z1 and ends at x2/y2/z2.
           The origin of the coordinate system is the left lower
           corner:


          z^
           |      ^y
           |     /
           |  +-----------+
           | /|/         /|nz
           |/ +---------/-+
           +-----------+ /ny
           |/          |/
           *-----------/-------->
                nx              x

           For each hit cell, its index i,j,k (0..nx-1, 0..ny-1,
           0..nz-1) and the pathlength of the ray inside the cell
           is returned. 

           The ray may start and/or end inside the volume, in which
           case the end points of the ray are projected onto the
           cube surface.

  Input:
    int    nx               Number of cells in x-direction
    double size_x           Size of a cell in x-direction
    int    ny               Number of cells in y-direction
    double size_y           Size of a cell in y-direction
    int    nz               Number of cells in z-direction
    double size_z           Size of a cell in z-direction

    double x1               Starting coordinate
    double y1
    double z1
    double x2               Ending coordinate
    double y2
    double z2

  Output:
    int    n_hit            number of hit cells
    int    array_i          array of i-indices for hit cells
    int    array_j          array of j-indices for hit cells
    int    array_k          array of k-indices for hit cells
    double array_pathlength array of pathlengs of ray inside hit cells

  Function value:
    none

\********************************************************************/
{
#define sqr(x) ((x)*(x))

double x, y, z;                                 /* current position */
double x_old, y_old, z_old;
double xc, yc, zc;
double pathlength;
double dist_x, dist_y, dist_z;
double dxy, dxz, dyx, dyz, dzx, dzy;
int    n;
int    ix, iy, iz, nix, niy, niz;
static double *xx, *xy, *xz, *yx, *yy, *yz, *zx, *zy, *zz;

  if (xx == NULL)
    {
    xx = malloc(sizeof(double) * (nx+1));
    xy = malloc(sizeof(double) * (nx+1));
    xz = malloc(sizeof(double) * (nx+1));
    yx = malloc(sizeof(double) * (ny+1));
    yy = malloc(sizeof(double) * (ny+1));
    yz = malloc(sizeof(double) * (ny+1));
    zx = malloc(sizeof(double) * (nz+1));
    zy = malloc(sizeof(double) * (nz+1));
    zz = malloc(sizeof(double) * (nz+1));
    }

  if (x1 == x2)
    x1 += size_x / 1E6;
  if (y1 == y2)
    y1 += size_y / 1E6;
  if (z1 == z2)
    z1 += size_z / 1E6;

  dxy = (x1 - x2) / (y1 - y2);
  dxz = (x1 - x2) / (z1 - z2);
  dyx = (y1 - y2) / (x1 - x2);
  dyz = (y1 - y2) / (z1 - z2);
  dzx = (z1 - z2) / (x1 - x2);
  dzy = (z1 - z2) / (y1 - y2);

  /* project the beginning and end of the ray onto the box */
  if (x1 < 0)
    {
    y1 = y1 + dyx*(0 - x1);
    z1 = z1 + dzx*(0 - x1);
    x1 = 0;
    y2 = y2 + dyx*(size_x*nx - x2);
    z2 = z2 + dzx*(size_x*nx - x2);
    x2 = size_x*nx;
    }
  if (x1 > size_x * nx)
    {
    y1 = y1 + dyx*(size_x*nx - x1);
    z1 = z1 + dzx*(size_x*nx - x1);
    x1 = size_x*nx;
    y2 = y2 + dyx*(0 - x2);
    z2 = z2 + dzx*(0 - x2);
    x2 = 0;
    }
  if (y1 < 0)
    {
    x1 = x1 + dxy*(0 - y1);
    z1 = z1 + dzy*(0 - y1);
    y1 = 0;
    x2 = x2 + dxy*(size_y*ny - y2);
    z2 = z2 + dzy*(size_y*ny - y2);
    y2 = size_y*ny;
    }
  if (y1 > size_y * ny)
    {
    x1 = x1 + dxy*(size_y*ny - y1);
    z1 = z1 + dzy*(size_y*ny - y1);
    y1 = size_y*ny;
    x2 = x2 + dxy*(0 - y2);
    z2 = z2 + dzy*(0 - y2);
    y2 = 0;
    }
  if (z1 < 0)
    {
    x1 = x1 + dxz*(0 - z1);
    y1 = y1 + dyz*(0 - z1);
    z1 = 0;
    x2 = x2 + dxz*(size_z*nz - z2);
    y2 = y2 + dyz*(size_z*nz - z2);
    z2 = size_z*nz;
    }
  if (z1 > size_z * nz)
    {
    x1 = x1 + dxz*(size_z*nz - z1);
    y1 = y1 + dyz*(size_z*nz - z1);
    z1 = size_z*nz;
    x2 = x2 + dxz*(0 - z2);
    y2 = y2 + dyz*(0 - z2);
    z2 = 0;
    }

  /* if ray is outside volume, return */
  if (x1 == x2 || y1 == y2 || z1 == z2)
    {
    *n_hit = 0;
    return;
    }

  /* calculate each plane crossing in x-direction */
  for (ix=0,nix=0 ; ; ix++)
    {
    if (x1 < x2)
      x = ((int) (x1 / size_x) + ix + 1) * size_x;
    else
      x = ((int) (x1 / size_x) - ix) * size_x;
    y = y2 - dyx*(x2 - x);
    z = z2 - dzx*(x2 - x);

    if (x1 > x2 && (x < 0 || x < x2))
      break;
    if (x1 < x2 && (x > size_x * nx || x > x2))
      break;
    if (y1 > y2 && (y < 0 || y < y2))
      break;
    if (y1 < y2 && (y > size_y * ny || y > y2))
      break;
    if (z1 > z2 && (z < 0 || z < z2))
      break;
    if (z1 < z2 && (z > size_z * nz || z > z2))
      break;

    if (x>=0 && x<=size_x*nx &&
        y>=0 && y<=size_y*ny &&
        z>=0 && z<=size_z*nz)
      {
      xx[nix] = x;
      xy[nix] = y;
      xz[nix] = z;
      nix++;
      }
    }

  /* calculate each plane crossing in y-direction */
  for (iy=0,niy=0 ; ; iy++)
    {
    if (y1 < y2)
      y = ((int) (y1 / size_y) + iy + 1) * size_y;
    else
      y = ((int) (y1 / size_y) - iy) * size_y;
    x = x2 - dxy*(y2 - y);
    z = z2 - dzy*(y2 - y);

    if (x1 > x2 && (x < 0 || x < x2))
      break;
    if (x1 < x2 && (x > size_x * nx || x > x2))
      break;
    if (y1 > y2 && (y < 0 || y < y2))
      break;
    if (y1 < y2 && (y > size_y * ny || y > y2))
      break;
    if (z1 > z2 && (z < 0 || z < z2))
      break;
    if (z1 < z2 && (z > size_z * nz || z > z2))
      break;

    if (x>=0 && x<=size_x*nx &&
        y>=0 && y<=size_y*ny &&
        z>=0 && z<=size_z*nz)
      {
      yx[niy] = x;
      yy[niy] = y;
      yz[niy] = z;
      niy++;
      }
    }

  /* calculate each plane crossing in z-direction */
  for (iz=0,niz=0 ; ; iz++)
    {
    if (z1 < z2)
      z = ((int) (z1 / size_z) + iz + 1) * size_z;
    else
      z = ((int) (z1 / size_z) - iz) * size_z;
    x = x2 - dxz*(z2 - z);
    y = y2 - dyz*(z2 - z);

    if (x1 > x2 && (x < 0 || x < x2))
      break;
    if (x1 < x2 && (x > size_x * nx || x > x2))
      break;
    if (y1 > y2 && (y < 0 || y < y2))
      break;
    if (y1 < y2 && (y > size_y * ny || y > y2))
      break;
    if (z1 > z2 && (z < 0 || z < z2))
      break;
    if (z1 < z2 && (z > size_z * nz || z > z2))
      break;

    if (x>=0 && x<=size_x*nx &&
        y>=0 && y<=size_y*ny &&
        z>=0 && z<=size_z*nz)
      {
      zx[niz] = x;
      zy[niz] = y;
      zz[niz] = z;
      niz++;
      }
    }

  /* now go through all these points */

  if (nix == 0 && niy == 0 && niz == 0)
    {
    *n_hit = 0;
    return;
    }

  n=0;
  ix = iy = iz = 0;

  /* first point */
  x = x1;
  y = y1;
  z = z1;

  x_old = x1;
  y_old = y1;
  z_old = z1;

  do
    {
    if (ix >= nix)
      dist_x = size_x * 1E6;
    else
      dist_x = sqr(x_old-xx[ix])+sqr(y_old-xy[ix])+sqr(z_old-xz[ix]);
    if (iy >= niy)
      dist_y = size_y * 1E6;
    else
      dist_y = sqr(x_old-yx[iy])+sqr(y_old-yy[iy])+sqr(z_old-yz[iy]);
    if (iz >= niz)
      dist_z = size_z * 1E6;
    else
      dist_z = sqr(x_old-zx[iz])+sqr(y_old-zy[iz])+sqr(z_old-zz[iz]);

    if (dist_x < dist_y && dist_x < dist_z)
      {
      pathlength = sqrt(dist_x);
      x = xx[ix];
      y = xy[ix];
      z = xz[ix];
      ix++;
      }

    if (dist_y <= dist_x && dist_y < dist_z)
      {
      pathlength = sqrt(dist_y);
      x = yx[iy];
      y = yy[iy];
      z = yz[iy];
      iy++;
      }

    if (dist_z <= dist_x && dist_z <= dist_y)
      {
      pathlength = sqrt(dist_z);
      x = zx[iz];
      y = zy[iz];
      z = zz[iz];
      iz++;
      }

    if (pathlength > size_x / 1E6)
      {
      xc = (x+x_old)/2;
      yc = (y+y_old)/2;
      zc = (z+z_old)/2;

      /* check if we are inside box */
      if (xc >= 0 && xc < size_x * nx &&
          yc >= 0 && yc < size_x * nx &&
          zc >= 0 && zc < size_x * nx)
        {
        /* inside -> store point */
        array_i[n] = (int) (xc / size_x);
        array_j[n] = (int) (yc / size_y);
        array_k[n] = (int) (zc / size_z);
        array_pathlength[n] = pathlength;
        n++;
        }
      }
    x_old = x;
    y_old = y;
    z_old = z;

    } while (ix < nix || iy < niy || iz < niz);

  *n_hit = n;
}
